@@ -182,7 +182,7 @@ std::string GetLegacySyntaxName(Edition edition) {
 
 Generator::Generator() : file_(nullptr) {}
 
-Generator::~Generator() {}
+Generator::~Generator() = default;
 
 GeneratorOptions Generator::ParseParameter(absl::string_view parameter,
                                            std::string* error) const {
@@ -361,13 +361,6 @@ void Generator::PrintTopBoilerplate() const {
                     "protobuf_python_version", PROTOBUF_PYTHON_VERSION_STRING);
   }
   printer_->Print("\"\"\"Generated protocol buffer code.\"\"\"\n");
-  if (!opensource_runtime_) {
-    // This import is needed so that compatibility proto1 compiler output
-    // inserted at protoc_insertion_point can refer to other protos like
-    // google3.a.b.c. Code generated by proto2 compiler doesn't do it, and
-    // instead uses aliases assigned when importing modules.
-    printer_->Print("import google3\n");
-  }
   bool runtime_version_disabled = false;
   printer_->Print(
       "from google.protobuf import descriptor as _descriptor\n"
@@ -486,8 +479,17 @@ std::string Generator::GetResolvedFeatures(
     // Assume these are all enums.  If we add non-enum global features or any
     // python-specific features, we will need to come back and improve this
     // logic.
-    ABSL_CHECK(field->enum_type() != nullptr)
-        << "Unexpected non-enum field found!";
+    if (field->type() != FieldDescriptor::TYPE_ENUM) {
+      ABSL_CHECK(field->is_extension())
+          << "Unsupported non-enum global feature found: "
+          << field->full_name();
+      // Placeholder for python-specific features.
+      ABSL_CHECK(field->number() != 1003)
+          << "Unsupported python-specific feature found: "
+          << field->full_name();
+      // Skip any non-python language-specific features.
+      continue;
+    }
     if (field->options().retention() == FieldOptions::RETENTION_SOURCE) {
       // Skip any source-retention features.
       continue;
@@ -569,8 +571,8 @@ void Generator::PrintResolvedFeatures() const {
 void Generator::PrintFileDescriptor() const {
   absl::flat_hash_map<absl::string_view, std::string> m;
   m["descriptor_name"] = kDescriptorKey;
-  m["name"] = file_->name();
-  m["package"] = file_->package();
+  m["name"] = std::string(file_->name());
+  m["package"] = std::string(file_->package());
   m["syntax"] = GetLegacySyntaxName(GetEdition(*file_));
   m["edition"] = Edition_Name(GetEdition(*file_));
   m["options"] = OptionsValue(proto_.options().SerializeAsString());
@@ -652,8 +654,8 @@ void Generator::PrintEnum(const EnumDescriptor& enum_descriptor,
   std::string module_level_descriptor_name =
       ModuleLevelDescriptorName(enum_descriptor);
   m["descriptor_name"] = module_level_descriptor_name;
-  m["name"] = enum_descriptor.name();
-  m["full_name"] = enum_descriptor.full_name();
+  m["name"] = std::string(enum_descriptor.name());
+  m["full_name"] = std::string(enum_descriptor.full_name());
   m["file"] = kDescriptorKey;
   const char enum_descriptor_template[] =
       "$descriptor_name$ = _descriptor.EnumDescriptor(\n"
@@ -708,6 +710,7 @@ void Generator::PrintMessageDescriptors() const {
   }
 }
 
+// TODO: Remove python service code from opensource.
 void Generator::PrintServiceDescriptors() const {
   for (int i = 0; i < file_->service_count(); ++i) {
     PrintServiceDescriptor(*file_->service(i));
@@ -726,7 +729,7 @@ void Generator::PrintServiceDescriptor(
     const ServiceDescriptor& descriptor) const {
   absl::flat_hash_map<absl::string_view, std::string> m;
   m["service_name"] = ModuleLevelServiceDescriptorName(descriptor);
-  m["name"] = descriptor.name();
+  m["name"] = std::string(descriptor.name());
   m["file"] = kDescriptorKey;
   printer_->Print(m, "$service_name$ = $file$.services_by_name['$name$']\n");
 }
@@ -776,8 +779,8 @@ void Generator::PrintServiceStub(const ServiceDescriptor& descriptor) const {
 void Generator::PrintDescriptor(const Descriptor& message_descriptor,
                                 const DescriptorProto& proto) const {
   absl::flat_hash_map<absl::string_view, std::string> m;
-  m["name"] = message_descriptor.name();
-  m["full_name"] = message_descriptor.full_name();
+  m["name"] = std::string(message_descriptor.name());
+  m["full_name"] = std::string(message_descriptor.full_name());
   m["file"] = kDescriptorKey;
 
   PrintNestedDescriptors(message_descriptor, proto);
@@ -813,7 +816,7 @@ void Generator::PrintDescriptor(const Descriptor& message_descriptor,
   for (int i = 0; i < message_descriptor.enum_type_count(); ++i) {
     const std::string descriptor_name =
         ModuleLevelDescriptorName(*message_descriptor.enum_type(i));
-    printer_->Print(descriptor_name.c_str());
+    printer_->Print(descriptor_name);
     printer_->Print(",\n");
   }
   printer_->Outdent();
@@ -842,8 +845,8 @@ void Generator::PrintDescriptor(const Descriptor& message_descriptor,
   for (int i = 0; i < message_descriptor.oneof_decl_count(); ++i) {
     const OneofDescriptor* desc = message_descriptor.oneof_decl(i);
     m.clear();
-    m["name"] = desc->name();
-    m["full_name"] = desc->full_name();
+    m["name"] = std::string(desc->name());
+    m["full_name"] = std::string(desc->full_name());
     m["index"] = absl::StrCat(desc->index());
     options_string =
         OptionsValue(proto.oneof_decl(i).options().SerializeAsString());
@@ -986,9 +989,9 @@ void Generator::FixForeignFieldsInDescriptor(
     absl::flat_hash_map<absl::string_view, std::string> m;
     const OneofDescriptor* oneof = descriptor.oneof_decl(i);
     m["descriptor_name"] = ModuleLevelDescriptorName(descriptor);
-    m["oneof_name"] = oneof->name();
+    m["oneof_name"] = std::string(oneof->name());
     for (int j = 0; j < oneof->field_count(); ++j) {
-      m["field_name"] = oneof->field(j)->name();
+      m["field_name"] = std::string(oneof->field(j)->name());
       printer_->Print(
           m,
           "$descriptor_name$.oneofs_by_name['$oneof_name$'].fields.append(\n"
@@ -1004,7 +1007,7 @@ void Generator::FixForeignFieldsInDescriptor(
 void Generator::AddMessageToFileDescriptor(const Descriptor& descriptor) const {
   absl::flat_hash_map<absl::string_view, std::string> m;
   m["descriptor_name"] = kDescriptorKey;
-  m["message_name"] = descriptor.name();
+  m["message_name"] = std::string(descriptor.name());
   m["message_descriptor_name"] = ModuleLevelDescriptorName(descriptor);
   const char file_descriptor_template[] =
       "$descriptor_name$.message_types_by_name['$message_name$'] = "
@@ -1016,7 +1019,7 @@ void Generator::AddServiceToFileDescriptor(
     const ServiceDescriptor& descriptor) const {
   absl::flat_hash_map<absl::string_view, std::string> m;
   m["descriptor_name"] = kDescriptorKey;
-  m["service_name"] = descriptor.name();
+  m["service_name"] = std::string(descriptor.name());
   m["service_descriptor_name"] = ModuleLevelServiceDescriptorName(descriptor);
   const char file_descriptor_template[] =
       "$descriptor_name$.services_by_name['$service_name$'] = "
@@ -1028,7 +1031,7 @@ void Generator::AddEnumToFileDescriptor(
     const EnumDescriptor& descriptor) const {
   absl::flat_hash_map<absl::string_view, std::string> m;
   m["descriptor_name"] = kDescriptorKey;
-  m["enum_name"] = descriptor.name();
+  m["enum_name"] = std::string(descriptor.name());
   m["enum_descriptor_name"] = ModuleLevelDescriptorName(descriptor);
   const char file_descriptor_template[] =
       "$descriptor_name$.enum_types_by_name['$enum_name$'] = "
@@ -1040,7 +1043,7 @@ void Generator::AddExtensionToFileDescriptor(
     const FieldDescriptor& descriptor) const {
   absl::flat_hash_map<absl::string_view, std::string> m;
   m["descriptor_name"] = kDescriptorKey;
-  m["field_name"] = descriptor.name();
+  m["field_name"] = std::string(descriptor.name());
   m["resolved_name"] = ResolveKeyword(descriptor.name());
   const char file_descriptor_template[] =
       "$descriptor_name$.extensions_by_name['$field_name$'] = "
@@ -1149,7 +1152,7 @@ void Generator::PrintEnumValueDescriptor(
   std::string options_string;
   proto.options().SerializeToString(&options_string);
   absl::flat_hash_map<absl::string_view, std::string> m;
-  m["name"] = descriptor.name();
+  m["name"] = std::string(descriptor.name());
   m["index"] = absl::StrCat(descriptor.index());
   m["number"] = absl::StrCat(descriptor.number());
   m["options"] = OptionsValue(options_string);
@@ -1167,13 +1170,12 @@ void Generator::PrintFieldDescriptor(const FieldDescriptor& field,
   std::string options_string;
   proto.options().SerializeToString(&options_string);
   absl::flat_hash_map<absl::string_view, std::string> m;
-  m["name"] = field.name();
-  m["full_name"] = field.full_name();
+  m["name"] = std::string(field.name());
+  m["full_name"] = std::string(field.full_name());
   m["index"] = absl::StrCat(field.index());
   m["number"] = absl::StrCat(field.number());
   m["type"] = absl::StrCat(field.type());
   m["cpp_type"] = absl::StrCat(field.cpp_type());
-  m["label"] = absl::StrCat(field.label());
   m["has_default_value"] = field.has_default_value() ? "True" : "False";
   m["default_value"] = StringifyDefaultValue(field);
   m["is_extension"] = field.is_extension() ? "True" : "False";
@@ -1181,6 +1183,13 @@ void Generator::PrintFieldDescriptor(const FieldDescriptor& field,
   m["json_name"] = field.has_json_name()
                        ? absl::StrCat(", json_name='", field.json_name(), "'")
                        : "";
+  if (field.is_required()) {
+    m["label"] = absl::StrCat(FieldDescriptor::Label::LABEL_REQUIRED);
+  } else if (field.is_repeated()) {
+    m["label"] = absl::StrCat(FieldDescriptor::Label::LABEL_REPEATED);
+  } else {
+    m["label"] = absl::StrCat(FieldDescriptor::Label::LABEL_OPTIONAL);
+  }
   // We always set message_type and enum_type to None at this point, and then
   // these fields in correctly after all referenced descriptors have been
   // defined and/or imported (see FixForeignFieldsInDescriptors()).
@@ -1462,7 +1471,7 @@ void Generator::FixOptionsForField(const FieldDescriptor& field,
   if (field.is_extension()) {
     if (field.extension_scope() == nullptr) {
       // Top level extensions.
-      field_name = field.name();
+      field_name = std::string(field.name());
     } else {
       field_name = FieldReferencingExpression(field.extension_scope(), field,
                                               "extensions_by_name");
