@@ -7,6 +7,8 @@
 
 #include "google/protobuf/compiler/python/pyi_generator.h"
 
+#include <cstddef>
+#include <memory>
 #include <string>
 #include <utility>
 #include <vector>
@@ -16,8 +18,10 @@
 #include "absl/log/absl_log.h"
 #include "absl/strings/ascii.h"
 #include "absl/strings/match.h"
+#include "absl/strings/str_cat.h"
 #include "absl/strings/str_split.h"
 #include "absl/strings/string_view.h"
+#include "absl/synchronization/mutex.h"
 #include "google/protobuf/compiler/code_generator.h"
 #include "google/protobuf/compiler/python/helpers.h"
 #include "google/protobuf/descriptor.h"
@@ -32,7 +36,7 @@ namespace python {
 
 PyiGenerator::PyiGenerator() : file_(nullptr) {}
 
-PyiGenerator::~PyiGenerator() {}
+PyiGenerator::~PyiGenerator() = default;
 
 template <typename DescriptorT>
 std::string PyiGenerator::ModuleLevelName(const DescriptorT& descriptor) const {
@@ -60,14 +64,16 @@ std::string PyiGenerator::InternalPackage() const {
 
 struct ImportModules {
   bool has_repeated = false;    // _containers
-  bool has_iterable = false;    // typing.Iterable
+  bool has_iterable = false;    // collections.abc.Iterable
   bool has_messages = false;    // _message
   bool has_enums = false;       // _enum_type_wrapper
   bool has_extendable = false;  // _python_message
-  bool has_mapping = false;     // typing.Mapping
+  bool has_mapping = false;     // collections.abc.Mapping
   bool has_optional = false;    // typing.Optional
   bool has_union = false;       // typing.Union
+  bool has_callable = false;    // typing.Callable
   bool has_well_known_type = false;
+  bool has_datetime = false;
 };
 
 // Checks whether a descriptor name matches a well-known type.
@@ -107,8 +113,15 @@ void CheckImportModules(const Descriptor* descriptor,
     if (field->is_map()) {
       import_modules->has_mapping = true;
       const FieldDescriptor* value_des = field->message_type()->field(1);
-      if (value_des->cpp_type() == FieldDescriptor::CPPTYPE_MESSAGE ||
-          value_des->cpp_type() == FieldDescriptor::CPPTYPE_ENUM) {
+      if (value_des->cpp_type() == FieldDescriptor::CPPTYPE_MESSAGE) {
+        import_modules->has_union = true;
+        const absl::string_view name = value_des->message_type()->full_name();
+        if (name == "google.protobuf.Duration" ||
+            name == "google.protobuf.Timestamp") {
+          import_modules->has_datetime = true;
+        }
+      }
+      if (value_des->cpp_type() == FieldDescriptor::CPPTYPE_ENUM) {
         import_modules->has_union = true;
       }
     } else {
@@ -118,6 +131,11 @@ void CheckImportModules(const Descriptor* descriptor,
       if (field->cpp_type() == FieldDescriptor::CPPTYPE_MESSAGE) {
         import_modules->has_union = true;
         import_modules->has_mapping = true;
+        const absl::string_view name = field->message_type()->full_name();
+        if (name == "google.protobuf.Duration" ||
+            name == "google.protobuf.Timestamp") {
+          import_modules->has_datetime = true;
+        }
       }
       if (field->cpp_type() == FieldDescriptor::CPPTYPE_ENUM) {
         import_modules->has_union = true;
@@ -165,6 +183,26 @@ void PyiGenerator::PrintImportForDescriptor(
 }
 
 void PyiGenerator::PrintImports() const {
+  // Checks what modules should be imported.
+  ImportModules import_modules;
+  if (file_->message_type_count() > 0) {
+    import_modules.has_messages = true;
+  }
+  if (file_->enum_type_count() > 0) {
+    import_modules.has_enums = true;
+  }
+  if (!opensource_runtime_ && file_->service_count() > 0) {
+    import_modules.has_callable = true;
+    import_modules.has_optional = true;
+    import_modules.has_union = true;
+  }
+  for (int i = 0; i < file_->message_type_count(); i++) {
+    CheckImportModules(file_->message_type(i), &import_modules);
+  }
+  if (import_modules.has_datetime) {
+    printer_->Print("import datetime\n\n");
+  }
+
   // Prints imported dependent _pb2 files.
   absl::flat_hash_set<std::string> seen_aliases;
   bool has_importlib = false;
@@ -180,22 +218,6 @@ void PyiGenerator::PrintImports() const {
     }
   }
 
-  // Checks what modules should be imported.
-  ImportModules import_modules;
-  if (file_->message_type_count() > 0) {
-    import_modules.has_messages = true;
-  }
-  if (file_->enum_type_count() > 0) {
-    import_modules.has_enums = true;
-  }
-  if (!opensource_runtime_ && file_->service_count() > 0) {
-    import_modules.has_optional = true;
-    import_modules.has_union = true;
-  }
-  for (int i = 0; i < file_->message_type_count(); i++) {
-    CheckImportModules(file_->message_type(i), &import_modules);
-  }
-
   // Prints modules (e.g. _containers, _messages, typing) that are
   // required in the proto file.
   if (import_modules.has_repeated) {
@@ -234,23 +256,35 @@ void PyiGenerator::PrintImports() const {
   } else {
     if (file_->service_count() > 0) {
       printer_->Print(
+          "from google3.net.rpc.python import legacy_stream_token as "
+          "_legacy_stream_token\n"
           "from google3.net.rpc.python import proto_python_api_2_stub as "
           "_proto_python_api_2_stub\n"
           "from google3.net.rpc.python import pywraprpc as _pywraprpc\n"
           "from google3.net.rpc.python import rpcserver as _rpcserver\n");
     }
   }
+  if (import_modules.has_iterable || import_modules.has_mapping) {
+    printer_->Print("from collections.abc import");
+    if (import_modules.has_iterable) {
+      printer_->Print(" Iterable as _Iterable");
+      if (import_modules.has_mapping) {
+        printer_->Print(",");
+      }
+    }
+    if (import_modules.has_mapping) {
+      printer_->Print(" Mapping as _Mapping");
+    }
+    printer_->Print("\n");
+  }
   printer_->Print("from typing import ");
   if (!opensource_runtime_ && file_->service_count() > 0) {
     printer_->Print("Any as _Any, ");
   }
+  if (import_modules.has_callable) {
+    printer_->Print("Callable as _Callable, ");
+  }
   printer_->Print("ClassVar as _ClassVar");
-  if (import_modules.has_iterable) {
-    printer_->Print(", Iterable as _Iterable");
-  }
-  if (import_modules.has_mapping) {
-    printer_->Print(", Mapping as _Mapping");
-  }
   if (import_modules.has_optional) {
     printer_->Print(", Optional as _Optional");
   }
@@ -276,7 +310,7 @@ void PyiGenerator::PrintImports() const {
                       public_dep->enum_type(i)->name());
     }
   }
-printer_->Print("\n");
+  printer_->Print("\n");
 }
 
 // Annotate wrapper for debugging purposes
@@ -284,7 +318,7 @@ printer_->Print("\n");
 template <typename DescriptorT>
 void PyiGenerator::Annotate(const std::string& label,
                             const DescriptorT* descriptor) const {
-printer_->Annotate(label.c_str(), descriptor);
+  printer_->Annotate(label, descriptor);
 }
 
 void PyiGenerator::PrintEnum(const EnumDescriptor& enum_descriptor) const {
@@ -380,8 +414,18 @@ std::string PyiGenerator::GetFieldType(
   return "";
 }
 
-void PyiGenerator::PrintMessage(
-    const Descriptor& message_descriptor, bool is_nested) const {
+std::string PyiGenerator::ExtraInitTypes(const Descriptor& msg_des) const {
+  if (msg_des.full_name() == "google.protobuf.Timestamp") {
+    return "datetime.datetime, ";
+  } else if (msg_des.full_name() == "google.protobuf.Duration") {
+    return "datetime.timedelta, ";
+  } else {
+    return "";
+  }
+}
+
+void PyiGenerator::PrintMessage(const Descriptor& message_descriptor,
+                                bool is_nested) const {
   if (!is_nested) {
     printer_->Print("\n");
   }
@@ -400,21 +444,7 @@ void PyiGenerator::PrintMessage(
   Annotate("class_name", &message_descriptor);
   printer_->Indent();
 
-  // Prints slots
-  printer_->Print("__slots__ = (");
-  int items_printed = 0;
-  for (int i = 0; i < message_descriptor.field_count(); ++i) {
-    const FieldDescriptor* field_des = message_descriptor.field(i);
-    if (IsPythonKeyword(field_des->name())) {
-      continue;
-    }
-    if (items_printed > 0) {
-      printer_->Print(", ");
-    }
-    ++items_printed;
-    printer_->Print("\"$field_name$\"", "field_name", field_des->name());
-  }
-  printer_->Print(items_printed == 1 ? ",)\n" : ")\n");
+  printer_->Print("__slots__ = ()\n");
 
   // Prints Extensions for extendable messages
   if (message_descriptor.extension_range_count() > 0) {
@@ -475,29 +505,28 @@ void PyiGenerator::PrintMessage(
   }
 
   // Prints __init__
-  printer_->Print("def __init__(self");
-  bool has_key_words = false;
-  bool is_first = true;
+  printer_->Print("def __init__(");
+  // If the message has a field named "self" (see b/144146793), it can still be
+  // passed to the initializer, which takes those as **kwargs. To avoid name
+  // collision, we rename the self parameter by appending underscores until it
+  // no longer collides. The self-parameter is in fact positional-only, so the
+  // name in the pyi doesn't matter with regard to what runtime usage is valid.
+  std::string self_arg_name = "self";
+  while (message_descriptor.FindFieldByName(self_arg_name) != nullptr) {
+    self_arg_name.append("_");
+  }
+  printer_->Print(self_arg_name);
+  bool has_python_keywords = false;
   for (int i = 0; i < message_descriptor.field_count(); ++i) {
     const FieldDescriptor* field_des = message_descriptor.field(i);
     if (IsPythonKeyword(field_des->name())) {
-      has_key_words = true;
+      has_python_keywords = true;
       continue;
     }
     std::string field_name = std::string(field_des->name());
-    if (is_first && field_name == "self") {
-      // See b/144146793 for an example of real code that generates a (self,
-      // self) method signature. Since repeating a parameter name is illegal in
-      // Python, we rename the duplicate self.
-      field_name = "self_";
-    }
-    is_first = false;
     printer_->Print(", $field_name$: ", "field_name", field_name);
     Annotate("field_name", field_des);
-    if (field_des->is_repeated() ||
-        field_des->cpp_type() != FieldDescriptor::CPPTYPE_BOOL) {
-      printer_->Print("_Optional[");
-    }
+    printer_->Print("_Optional[");
     if (field_des->is_map()) {
       const Descriptor* map_entry = field_des->message_type();
       printer_->Print(
@@ -510,9 +539,11 @@ void PyiGenerator::PrintMessage(
         printer_->Print("_Iterable[");
       }
       if (field_des->cpp_type() == FieldDescriptor::CPPTYPE_MESSAGE) {
-        printer_->Print(
-            "_Union[$type_name$, _Mapping]", "type_name",
-            GetFieldType(*field_des, message_descriptor));
+        const auto& extra_init_types =
+            ExtraInitTypes(*field_des->message_type());
+        printer_->Print("_Union[$extra_init_types$$type_name$, _Mapping]",
+                        "extra_init_types", extra_init_types, "type_name",
+                        GetFieldType(*field_des, message_descriptor));
       } else {
         if (field_des->cpp_type() == FieldDescriptor::CPPTYPE_ENUM) {
           printer_->Print("_Union[$type_name$, str]", "type_name",
@@ -527,13 +558,9 @@ void PyiGenerator::PrintMessage(
         printer_->Print("]");
       }
     }
-    if (field_des->is_repeated() ||
-        field_des->cpp_type() != FieldDescriptor::CPPTYPE_BOOL) {
-      printer_->Print("]");
-    }
-    printer_->Print(" = ...");
+    printer_->Print("] = ...");
   }
-  if (has_key_words) {
+  if (has_python_keywords) {
     printer_->Print(", **kwargs");
   }
   printer_->Print(") -> None: ...\n");
