@@ -12,9 +12,11 @@
 #include <cstdint>
 #include <cstdlib>
 #include <cstring>
+#include <functional>
 #include <iterator>
 #include <list>
 #include <memory>
+#include <optional>
 #include <string>
 #include <type_traits>
 #include <utility>
@@ -25,10 +27,13 @@
 #include "absl/log/absl_check.h"
 #include "absl/numeric/bits.h"
 #include "absl/strings/str_cat.h"
+#include "absl/types/span.h"
 #include "google/protobuf/arena_test_util.h"
 #include "google/protobuf/internal_visibility_for_testing.h"
 #include "google/protobuf/io/coded_stream.h"
+#include "google/protobuf/message.h"
 #include "google/protobuf/unittest.pb.h"
+#include "google/protobuf/unittest_import.pb.h"
 
 
 // Must be included last.
@@ -36,17 +41,27 @@
 
 namespace google {
 namespace protobuf {
-namespace {
+namespace internal {
 
-using ::protobuf_unittest::TestAllTypes;
-using ::protobuf_unittest::TestMessageWithManyRepeatedPtrFields;
+using ::proto2_unittest::TestAllTypes;
+using ::proto2_unittest::TestMessageWithManyRepeatedPtrFields;
 using ::testing::A;
 using ::testing::AllOf;
 using ::testing::ElementsAre;
 using ::testing::Ge;
 using ::testing::Le;
 
-TEST(RepeatedPtrOverPtrsIterator, Traits) {
+using String = std::string;
+using SmallMsg = proto2_unittest::ForeignMessage;
+using LargeMsg = proto2_unittest::TestAllTypes;
+
+namespace {
+enum WithArena : bool { kNoArena = false, kArena = true };
+enum AddMode { kIntoNew = 0, kIntoCleared = 1 };
+enum RefMode { kConcrete = 0, kAbstract = 1 };
+}  // namespace
+
+TEST(RepeatedPtrOverPtrsIteratorTest, Traits) {
   using It = RepeatedPtrField<std::string>::pointer_iterator;
   static_assert(std::is_same<It::value_type, std::string*>::value, "");
   static_assert(std::is_same<It::reference, std::string*&>::value, "");
@@ -55,7 +70,7 @@ TEST(RepeatedPtrOverPtrsIterator, Traits) {
   static_assert(std::is_same<It::iterator_category,
                              std::random_access_iterator_tag>::value,
                 "");
-#if __cplusplus >= 202002L
+#if PROTOBUF_CPLUSPLUS_MIN(202002L)
   static_assert(
       std::is_same<It::iterator_concept, std::contiguous_iterator_tag>::value,
       "");
@@ -67,7 +82,7 @@ TEST(RepeatedPtrOverPtrsIterator, Traits) {
 }
 
 #ifdef __cpp_lib_to_address
-TEST(RepeatedPtrOverPtrsIterator, ToAddress) {
+TEST(RepeatedPtrOverPtrsIteratorTest, ToAddress) {
   // empty container
   RepeatedPtrField<std::string> field;
   EXPECT_THAT(std::to_address(field.pointer_begin()), A<std::string**>());
@@ -91,7 +106,7 @@ TEST(ConstRepeatedPtrOverPtrsIterator, Traits) {
   static_assert(std::is_same<It::iterator_category,
                              std::random_access_iterator_tag>::value,
                 "");
-#if __cplusplus >= 202002L
+#if PROTOBUF_CPLUSPLUS_MIN(202002L)
   static_assert(
       std::is_same<It::iterator_concept, std::contiguous_iterator_tag>::value,
       "");
@@ -102,7 +117,44 @@ TEST(ConstRepeatedPtrOverPtrsIterator, Traits) {
 #endif
 }
 
-TEST(RepeatedPtrField, MoveAdd) {
+TEST(RepeatedPtrFieldTest, SimpleAddWithStrings) {
+  RepeatedPtrField<std::string> field;
+  ASSERT_EQ(field.size(), 0);
+  field.Add("foo");
+  ASSERT_EQ(field.size(), 1);
+  field.Add("bar");
+  ASSERT_EQ(field.size(), 2);
+  field.Add("buz");
+  ASSERT_EQ(field.size(), 3);
+  field.Add("qux");
+  ASSERT_EQ(field.size(), 4);
+  EXPECT_EQ(field.Get(0), "foo");
+  EXPECT_EQ(field.Get(1), "bar");
+  EXPECT_EQ(field.Get(2), "buz");
+  EXPECT_EQ(field.Get(3), "qux");
+}
+
+TEST(RepeatedPtrFieldTest, SimpleAddWithMessages) {
+  using TestType = TestAllTypes::NestedMessage;
+  RepeatedPtrField<TestType> field;
+  const auto make_val = [](int i) {
+    TestType x;
+    x.set_bb(i);
+    return x;
+  };
+  ASSERT_EQ(field.size(), 0);
+  field.Add(make_val(1));
+  ASSERT_EQ(field.size(), 1);
+  field.Add(make_val(2));
+  ASSERT_EQ(field.size(), 2);
+  field.Add(make_val(3));
+  ASSERT_EQ(field.size(), 3);
+  EXPECT_EQ(field.Get(0).bb(), 1);
+  EXPECT_EQ(field.Get(1).bb(), 2);
+  EXPECT_EQ(field.Get(2).bb(), 3);
+}
+
+TEST(RepeatedPtrFieldTest, MoveAdd) {
   RepeatedPtrField<TestAllTypes> field;
   TestAllTypes test_all_types;
   auto* optional_nested_message =
@@ -114,12 +166,12 @@ TEST(RepeatedPtrField, MoveAdd) {
             field.Mutable(0)->mutable_optional_nested_message());
 }
 
-TEST(RepeatedPtrField, ConstInit) {
+TEST(RepeatedPtrFieldTest, ConstInit) {
   PROTOBUF_CONSTINIT static RepeatedPtrField<std::string> field{};  // NOLINT
   EXPECT_TRUE(field.empty());
 }
 
-TEST(RepeatedPtrField, ClearThenReserveMore) {
+TEST(RepeatedPtrFieldTest, ClearThenReserveMore) {
   // Test that Reserve properly destroys the old internal array when it's forced
   // to allocate a new one, even when cleared-but-not-deleted objects are
   // present. Use a 'string' and > 16 bytes length so that the elements are
@@ -127,7 +179,7 @@ TEST(RepeatedPtrField, ClearThenReserveMore) {
   // calls here.
   RepeatedPtrField<std::string> field;
   for (int i = 0; i < 32; i++) {
-    *field.Add() = std::string("abcdefghijklmnopqrstuvwxyz0123456789");
+    *field.Add() = "abcdefghijklmnopqrstuvwxyz0123456789";
   }
   EXPECT_EQ(32, field.size());
   field.Clear();
@@ -150,7 +202,7 @@ auto ValidResolutionPointerRange(const std::string* p)
 template <typename X>
 std::false_type ValidResolutionPointerRange(void*);
 
-TEST(RepeatedPtrField, UnambiguousConstructor) {
+TEST(RepeatedPtrFieldTest, UnambiguousConstructor) {
   struct X {
     static bool f(std::vector<std::string>) { return false; }
     static bool f(google::protobuf::RepeatedPtrField<std::string>) { return true; }
@@ -180,7 +232,7 @@ TEST(RepeatedPtrField, UnambiguousConstructor) {
   EXPECT_FALSE(decltype(ValidResolutionPointerRange<X>(nullptr))::value);
 }
 
-TEST(RepeatedPtrField, Small) {
+TEST(RepeatedPtrFieldTest, Small) {
   RepeatedPtrField<std::string> field;
 
   EXPECT_TRUE(field.empty());
@@ -224,7 +276,7 @@ TEST(RepeatedPtrField, Small) {
   EXPECT_EQ(field.size(), 0);
 }
 
-TEST(RepeatedPtrField, Large) {
+TEST(RepeatedPtrFieldTest, Large) {
   RepeatedPtrField<std::string> field;
 
   for (int i = 0; i < 16; i++) {
@@ -242,24 +294,25 @@ TEST(RepeatedPtrField, Large) {
   EXPECT_GE(field.SpaceUsedExcludingSelf(), min_expected_usage);
 }
 
-// TODO: Dedup with the same thing in repeated_field_unittest.cc.
-template <typename Rep>
+namespace {
+
+template <typename Elem>
 void CheckAllocationSizes() {
-  using T = typename Rep::value_type;
+  using Field = RepeatedPtrField<Elem>;
   // Use a large initial block to make the checks below easier to predict.
   std::string buf(1 << 20, 0);
   Arena arena(&buf[0], buf.size());
-  auto* rep = Arena::Create<Rep>(&arena);
-  size_t prev = arena.SpaceUsed();
+  auto* field = Arena::Create<Field>(&arena);
+  size_t prev_used = arena.SpaceUsed();
 
   for (int i = 0; i < 100; ++i) {
-    rep->Add(T{});
+    field->Add(Elem{});
     if (sizeof(void*) == 8) {
       // For `RepeatedPtrField`, we also allocate the T on the arena.
       // Subtract those from the count.
-      size_t new_used = arena.SpaceUsed() - (sizeof(T) * (i + 1));
-      size_t last_alloc = new_used - prev;
-      prev = new_used;
+      size_t new_used = arena.SpaceUsed() - (sizeof(Elem) * (i + 1));
+      size_t last_alloc = new_used - prev_used;
+      prev_used = new_used;
 
       // When we actually allocated something, check the size.
       if (last_alloc != 0) {
@@ -271,7 +324,7 @@ void CheckAllocationSizes() {
       }
 
       // The byte size must be a multiple of 8 when not SOO.
-      const int capacity_bytes = rep->Capacity() * sizeof(T);
+      const int capacity_bytes = field->Capacity() * sizeof(Elem);
       if (capacity_bytes > internal::kSooCapacityBytes) {
         ASSERT_EQ(capacity_bytes % 8, 0);
       }
@@ -279,36 +332,44 @@ void CheckAllocationSizes() {
   }
 }
 
-TEST(RepeatedPtrField, ArenaAllocationSizesMatchExpectedValues) {
-  EXPECT_NO_FATAL_FAILURE(
-      CheckAllocationSizes<RepeatedPtrField<std::string>>());
+}  // namespace
+
+TEST(RepeatedPtrFieldTest, ArenaAllocationSizesMatchExpectedValues) {
+  EXPECT_NO_FATAL_FAILURE(CheckAllocationSizes<std::string>());
+  EXPECT_NO_FATAL_FAILURE(CheckAllocationSizes<TestAllTypes::NestedMessage>());
 }
 
-TEST(RepeatedPtrField, NaturalGrowthOnArenasReuseBlocks) {
-  using Rep = RepeatedPtrField<std::string>;
-  Arena arena;
-  std::vector<Rep*> values;
+TEST(RepeatedPtrFieldTest, NaturalGrowthOnArenasReuseBlocks) {
+  using Elem = std::string;
+  using Field = RepeatedPtrField<Elem>;
 
+  Arena arena;
+  std::vector<Field*> fields;
   static constexpr int kNumFields = 100;
   static constexpr int kNumElems = 1000;
+  std::optional<int> common_capacity;
   for (int i = 0; i < kNumFields; ++i) {
-    values.push_back(Arena::Create<Rep>(&arena));
-    auto& field = *values.back();
+    fields.push_back(Arena::Create<Field>(&arena));
+    auto& field = *fields.back();
     for (int j = 0; j < kNumElems; ++j) {
       field.Add("");
     }
+    if (!common_capacity.has_value()) {
+      common_capacity = field.Capacity();
+    } else {
+      ASSERT_EQ(field.Capacity(), *common_capacity);
+    }
   }
 
-  size_t expected =
-      values.size() * values[0]->Capacity() * sizeof(std::string*) +
-      sizeof(std::string) * kNumElems * kNumFields;
+  const size_t expected = kNumFields * (*common_capacity) * sizeof(Elem*) +
+                          kNumFields * kNumElems * sizeof(Elem);
   // Use a 2% slack for other overhead.
   // If we were not reusing the blocks, the actual value would be ~2x the
   // expected.
   EXPECT_THAT(arena.SpaceUsed(), AllOf(Ge(expected), Le(1.02 * expected)));
 }
 
-TEST(RepeatedPtrField, AddAndAssignRanges) {
+TEST(RepeatedPtrFieldTest, AddAndAssignRanges) {
   RepeatedPtrField<std::string> field;
 
   const char* vals[] = {"abc", "x", "yz", "xyzzy"};
@@ -332,7 +393,8 @@ TEST(RepeatedPtrField, AddAndAssignRanges) {
   EXPECT_EQ(field.Get(7), "xyzzy");
 }
 
-TEST(RepeatedPtrField, SwapSmallSmall) {
+
+TEST(RepeatedPtrFieldTest, SwapSmallSmall) {
   RepeatedPtrField<std::string> field1;
   RepeatedPtrField<std::string> field2;
 
@@ -362,7 +424,7 @@ TEST(RepeatedPtrField, SwapSmallSmall) {
   EXPECT_EQ(field2.Get(1), "bar");
 }
 
-TEST(RepeatedPtrField, SwapLargeSmall) {
+TEST(RepeatedPtrFieldTest, SwapLargeSmall) {
   RepeatedPtrField<std::string> field1;
   RepeatedPtrField<std::string> field2;
 
@@ -383,7 +445,7 @@ TEST(RepeatedPtrField, SwapLargeSmall) {
   }
 }
 
-TEST(RepeatedPtrField, SwapLargeLarge) {
+TEST(RepeatedPtrFieldTest, SwapLargeLarge) {
   RepeatedPtrField<std::string> field1;
   RepeatedPtrField<std::string> field2;
 
@@ -418,14 +480,14 @@ static int ReservedSpace(RepeatedPtrField<std::string>* field) {
   return field->size() - 1;
 }
 
-TEST(RepeatedPtrField, ReserveMoreThanDouble) {
+TEST(RepeatedPtrFieldTest, ReserveMoreThanDouble) {
   RepeatedPtrField<std::string> field;
   field.Reserve(20);
 
   EXPECT_LE(20, ReservedSpace(&field));
 }
 
-TEST(RepeatedPtrField, ReserveLessThanDouble) {
+TEST(RepeatedPtrFieldTest, ReserveLessThanDouble) {
   RepeatedPtrField<std::string> field;
   field.Reserve(20);
 
@@ -436,7 +498,7 @@ TEST(RepeatedPtrField, ReserveLessThanDouble) {
   EXPECT_LE(2 * capacity, ReservedSpace(&field));
 }
 
-TEST(RepeatedPtrField, ReserveLessThanExisting) {
+TEST(RepeatedPtrFieldTest, ReserveLessThanExisting) {
   RepeatedPtrField<std::string> field;
   field.Reserve(20);
   const std::string* const* previous_ptr = field.data();
@@ -446,7 +508,7 @@ TEST(RepeatedPtrField, ReserveLessThanExisting) {
   EXPECT_LE(20, ReservedSpace(&field));
 }
 
-TEST(RepeatedPtrField, ReserveDoesntLoseAllocated) {
+TEST(RepeatedPtrFieldTest, ReserveDoesntLoseAllocated) {
   // Check that a bug is fixed:  An earlier implementation of Reserve()
   // failed to copy pointers to allocated-but-cleared objects, possibly
   // leading to segfaults.
@@ -458,40 +520,36 @@ TEST(RepeatedPtrField, ReserveDoesntLoseAllocated) {
   EXPECT_EQ(first, field.Add());
 }
 
-// Clearing elements is tricky with RepeatedPtrFields since the memory for
-// the elements is retained and reused.
-TEST(RepeatedPtrField, ClearedElements) {
-  PROTOBUF_IGNORE_DEPRECATION_START
-  RepeatedPtrField<std::string> field;
 
-  std::string* original = field.Add();
-  *original = "foo";
-
-  EXPECT_EQ(field.ClearedCount(), 0);
-
-  field.RemoveLast();
-  EXPECT_TRUE(original->empty());
-  EXPECT_EQ(field.ClearedCount(), 1);
-
-  EXPECT_EQ(field.Add(),
-            original);  // Should return same string for reuse.
-  EXPECT_EQ(field.UnsafeArenaReleaseLast(), original);  // We take ownership.
-  EXPECT_EQ(field.ClearedCount(), 0);
-
-  EXPECT_NE(field.Add(), original);  // Should NOT return the same string.
-  EXPECT_EQ(field.ClearedCount(), 0);
-
-  field.UnsafeArenaAddAllocated(original);  // Give ownership back.
-  EXPECT_EQ(field.ClearedCount(), 0);
-  EXPECT_EQ(field.Mutable(1), original);
-
-  field.Clear();
-  EXPECT_EQ(field.ClearedCount(), 2);
-  PROTOBUF_IGNORE_DEPRECATION_STOP
+// TODO: Re-evaluate if this is still needed once the bug is fixed.
+TEST(RepeatedPtrFieldTest, AddRvalueToCleared) {
+  // Check that an added rvalue correctly overwrites a cleared SOO element.
+  {
+    RepeatedPtrField<std::string> field;
+    field.Add()->assign("foo");
+    ASSERT_THAT(field, ElementsAre("foo"));
+    field.RemoveLast();
+    ASSERT_EQ(field.size(), 0);
+    field.Add(std::string{"bar"});
+    EXPECT_THAT(field, ElementsAre("bar"));
+  }
+  // Check that an added rvalue correctly overwrites a cleared non-SOO element
+  // in the Rep.
+  {
+    RepeatedPtrField<std::string> field;
+    field.Add()->assign("foo");
+    field.Add()->assign("bar");
+    field.Add()->assign("baz");
+    EXPECT_THAT(field, ElementsAre("foo", "bar", "baz"));
+    field.RemoveLast();
+    EXPECT_THAT(field, ElementsAre("foo", "bar"));
+    field.Add(std::string{"qux"});
+    EXPECT_THAT(field, ElementsAre("foo", "bar", "qux"));
+  }
 }
 
 // Test all code paths in AddAllocated().
-TEST(RepeatedPtrField, AddAllocated) {
+TEST(RepeatedPtrFieldTest, AddAllocated) {
   RepeatedPtrField<std::string> field;
   while (field.size() < field.Capacity()) {
     field.Add()->assign("filler");
@@ -512,7 +570,6 @@ TEST(RepeatedPtrField, AddAllocated) {
   std::string* foo = new std::string("foo");
   field.AddAllocated(foo);
   EXPECT_EQ(index + 1, field.size());
-  EXPECT_EQ(0, field.ClearedCount());
   EXPECT_EQ(foo, &field.Get(index));
 
   // Last branch:  Field is not at capacity and there are no cleared objects.
@@ -521,7 +578,6 @@ TEST(RepeatedPtrField, AddAllocated) {
   field.AddAllocated(bar);
   ++index;
   EXPECT_EQ(index + 1, field.size());
-  EXPECT_EQ(0, field.ClearedCount());
   EXPECT_EQ(bar, &field.Get(index));
 
   // Third branch:  Field is not at capacity and there are no cleared objects.
@@ -530,7 +586,6 @@ TEST(RepeatedPtrField, AddAllocated) {
   std::string* baz = new std::string("baz");
   field.AddAllocated(baz);
   EXPECT_EQ(index + 1, field.size());
-  EXPECT_EQ(1, field.ClearedCount());
   EXPECT_EQ(baz, &field.Get(index));
 
   // Second branch:  Field is at capacity but has some cleared objects.
@@ -541,11 +596,10 @@ TEST(RepeatedPtrField, AddAllocated) {
   field.AddAllocated(moo);
   EXPECT_EQ(index + 1, field.size());
   // We should have discarded the cleared object.
-  EXPECT_EQ(0, field.ClearedCount());
   EXPECT_EQ(moo, &field.Get(index));
 }
 
-TEST(RepeatedPtrField, AddMethodsDontAcceptNull) {
+TEST(RepeatedPtrFieldDeathTest, AddMethodsDontAcceptNull) {
 #if !defined(NDEBUG)
   RepeatedPtrField<std::string> field;
   EXPECT_DEATH(field.AddAllocated(nullptr), "nullptr");
@@ -553,14 +607,40 @@ TEST(RepeatedPtrField, AddMethodsDontAcceptNull) {
 #endif
 }
 
-TEST(RepeatedPtrField, AddAllocatedDifferentArena) {
+TEST(RepeatedPtrFieldTest, AddAllocatedDifferentArena) {
   RepeatedPtrField<TestAllTypes> field;
   Arena arena;
   auto* msg = Arena::Create<TestAllTypes>(&arena);
   field.AddAllocated(msg);
 }
 
-TEST(RepeatedPtrField, MergeFrom) {
+// This test replicates a very specific scenario that used to cause a transient
+// hard-to-debug failure in protoc during development.
+TEST(RepeatedPtrFieldTest, UnsafeArenaAddAllocatedReleaseLastOnBaseField) {
+  using ElemT = TestAllTypes::NestedMessage;
+  using FieldT = RepeatedPtrField<ElemT>;
+  Arena arena;
+  auto* concrete_field = Arena::Create<FieldT>(&arena);
+  ElemT* concrete_elem = concrete_field->Add();
+  concrete_elem->set_bb(123);
+  auto* base_field = reinterpret_cast<RepeatedPtrFieldBase*>(concrete_field);
+  const Message& base_prototype =
+      base_field->Get<GenericTypeHandler<Message>>(0);
+  Message* base_new_elem = base_prototype.New(&arena);
+  ASSERT_NE(base_new_elem, nullptr);
+  ElemT* concrete_new_elem = static_cast<ElemT*>(base_new_elem);
+  concrete_new_elem->set_bb(456);
+  base_field->UnsafeArenaAddAllocated<GenericTypeHandler<Message>>(
+      &arena, base_new_elem);
+  Message* base_new_elem_roundtrip =
+      base_field->UnsafeArenaReleaseLast<GenericTypeHandler<Message>>();
+  ASSERT_NE(base_new_elem_roundtrip, nullptr);
+  ElemT* concrete_new_elem_roundtrip =
+      static_cast<ElemT*>(base_new_elem_roundtrip);
+  EXPECT_EQ(concrete_new_elem_roundtrip->bb(), 456);
+}
+
+TEST(RepeatedPtrFieldTest, MergeFromString) {
   RepeatedPtrField<std::string> source, destination;
   source.Add()->assign("4");
   source.Add()->assign("5");
@@ -576,10 +656,114 @@ TEST(RepeatedPtrField, MergeFrom) {
   EXPECT_EQ("3", destination.Get(2));
   EXPECT_EQ("4", destination.Get(3));
   EXPECT_EQ("5", destination.Get(4));
+
+  destination.Clear();
+
+  ASSERT_EQ(0, destination.size());
+
+  destination.MergeFrom(source);
+
+  ASSERT_EQ(2, destination.size());
+  EXPECT_EQ("4", destination.Get(0));
+  EXPECT_EQ("5", destination.Get(1));
+
+  delete destination.ReleaseLast();
+
+  ASSERT_EQ(1, destination.size());
+  EXPECT_EQ("4", destination.Get(0));
+
+  destination.MergeFrom(source);
+
+  ASSERT_EQ(3, destination.size());
+  EXPECT_EQ("4", destination.Get(0));
+  EXPECT_EQ("4", destination.Get(1));
+  EXPECT_EQ("5", destination.Get(2));
+}
+
+TEST(RepeatedPtrFieldTest, MergeFromMessage) {
+  RepeatedPtrField<TestAllTypes::NestedMessage> source, destination;
+  source.Add()->set_bb(4);
+  source.Add()->set_bb(5);
+  destination.Add()->set_bb(1);
+  destination.Add()->set_bb(2);
+  destination.Add()->set_bb(3);
+
+  destination.MergeFrom(source);
+
+  ASSERT_EQ(5, destination.size());
+  EXPECT_EQ(1, destination.Get(0).bb());
+  EXPECT_EQ(2, destination.Get(1).bb());
+  EXPECT_EQ(3, destination.Get(2).bb());
+  EXPECT_EQ(4, destination.Get(3).bb());
+  EXPECT_EQ(5, destination.Get(4).bb());
+
+  destination.Clear();
+
+  ASSERT_EQ(0, destination.size());
+
+  destination.MergeFrom(source);
+
+  ASSERT_EQ(2, destination.size());
+  EXPECT_EQ(4, destination.Get(0).bb());
+  EXPECT_EQ(5, destination.Get(1).bb());
+
+  delete destination.ReleaseLast();
+
+  ASSERT_EQ(1, destination.size());
+  EXPECT_EQ(4, destination.Get(0).bb());
+
+  destination.MergeFrom(source);
+
+  ASSERT_EQ(3, destination.size());
+  EXPECT_EQ(4, destination.Get(0).bb());
+  EXPECT_EQ(4, destination.Get(1).bb());
+  EXPECT_EQ(5, destination.Get(2).bb());
+}
+
+TEST(RepeatedPtrFieldTest, MergeFromStringWithArena) {
+  using Field = RepeatedPtrField<std::string>;
+  Arena arena;
+  Field* source = Arena::Create<Field>(&arena);
+  Field* destination = Arena::Create<Field>(&arena);
+  source->Add()->assign("4");
+  source->Add()->assign("5");
+  destination->Add()->assign("1");
+  destination->Add()->assign("2");
+  destination->Add()->assign("3");
+
+  destination->MergeFrom(*source);
+
+  ASSERT_EQ(5, destination->size());
+  EXPECT_EQ("1", destination->Get(0));
+  EXPECT_EQ("2", destination->Get(1));
+  EXPECT_EQ("3", destination->Get(2));
+  EXPECT_EQ("4", destination->Get(3));
+  EXPECT_EQ("5", destination->Get(4));
+}
+
+TEST(RepeatedPtrFieldTest, MergeFromMessageWithArena) {
+  using Field = RepeatedPtrField<TestAllTypes::NestedMessage>;
+  Arena arena;
+  Field* source = Arena::Create<Field>(&arena);
+  Field* destination = Arena::Create<Field>(&arena);
+  source->Add()->set_bb(4);
+  source->Add()->set_bb(5);
+  destination->Add()->set_bb(1);
+  destination->Add()->set_bb(2);
+  destination->Add()->set_bb(3);
+
+  destination->MergeFrom(*source);
+
+  ASSERT_EQ(5, destination->size());
+  EXPECT_EQ(1, destination->Get(0).bb());
+  EXPECT_EQ(2, destination->Get(1).bb());
+  EXPECT_EQ(3, destination->Get(2).bb());
+  EXPECT_EQ(4, destination->Get(3).bb());
+  EXPECT_EQ(5, destination->Get(4).bb());
 }
 
 
-TEST(RepeatedPtrField, CopyFrom) {
+TEST(RepeatedPtrFieldTest, CopyFrom) {
   RepeatedPtrField<std::string> source, destination;
   source.Add()->assign("4");
   source.Add()->assign("5");
@@ -594,7 +778,7 @@ TEST(RepeatedPtrField, CopyFrom) {
   EXPECT_EQ("5", destination.Get(1));
 }
 
-TEST(RepeatedPtrField, CopyFromSelf) {
+TEST(RepeatedPtrFieldTest, CopyFromSelf) {
   RepeatedPtrField<std::string> me;
   me.Add()->assign("1");
   me.CopyFrom(me);
@@ -602,7 +786,7 @@ TEST(RepeatedPtrField, CopyFromSelf) {
   EXPECT_EQ("1", me.Get(0));
 }
 
-TEST(RepeatedPtrField, Erase) {
+TEST(RepeatedPtrFieldTest, Erase) {
   RepeatedPtrField<std::string> me;
   RepeatedPtrField<std::string>::iterator it = me.erase(me.begin(), me.end());
   EXPECT_TRUE(me.begin() == it);
@@ -635,7 +819,7 @@ TEST(RepeatedPtrField, Erase) {
   EXPECT_EQ("8", me.Get(2));
 }
 
-TEST(RepeatedPtrField, CopyConstruct) {
+TEST(RepeatedPtrFieldTest, CopyConstruct) {
   auto token = internal::InternalVisibilityForTesting{};
   RepeatedPtrField<std::string> source;
   source.Add()->assign("1");
@@ -652,7 +836,7 @@ TEST(RepeatedPtrField, CopyConstruct) {
   EXPECT_EQ("2", destination2.Get(1));
 }
 
-TEST(RepeatedPtrField, CopyConstructWithArena) {
+TEST(RepeatedPtrFieldTest, CopyConstructWithArena) {
   auto token = internal::InternalVisibilityForTesting{};
   RepeatedPtrField<std::string> source;
   source.Add()->assign("1");
@@ -665,7 +849,7 @@ TEST(RepeatedPtrField, CopyConstructWithArena) {
   EXPECT_EQ("2", destination.Get(1));
 }
 
-TEST(RepeatedPtrField, IteratorConstruct_String) {
+TEST(RepeatedPtrFieldTest, IteratorConstruct_String) {
   std::vector<std::string> values;
   values.push_back("1");
   values.push_back("2");
@@ -681,7 +865,7 @@ TEST(RepeatedPtrField, IteratorConstruct_String) {
   EXPECT_EQ(values[1], other.Get(1));
 }
 
-TEST(RepeatedPtrField, IteratorConstruct_Proto) {
+TEST(RepeatedPtrFieldTest, IteratorConstruct_Proto) {
   typedef TestAllTypes::NestedMessage Nested;
   std::vector<Nested> values;
   values.push_back(Nested());
@@ -700,7 +884,7 @@ TEST(RepeatedPtrField, IteratorConstruct_Proto) {
   EXPECT_EQ(values[1].bb(), other.Get(1).bb());
 }
 
-TEST(RepeatedPtrField, SmallOptimization) {
+TEST(RepeatedPtrFieldTest, SmallOptimization) {
   // Properties checked here are not part of the contract of RepeatedPtrField,
   // but we test them to verify that SSO is working as expected by the
   // implementation.
@@ -733,17 +917,21 @@ TEST(RepeatedPtrField, SmallOptimization) {
   // Adding a second object stops sso.
   std::string str2;
   array->UnsafeArenaAddAllocated(&str2);
-  EXPECT_EQ(array->Capacity(), 3);
-  // Backing array and the strings.
-  EXPECT_EQ(array->SpaceUsedExcludingSelf(),
-            (1 + array->Capacity()) * sizeof(void*) + 2 * sizeof(str));
+  // We know we have exited sso if the capacity is greater than 1.
+  EXPECT_GT(array->Capacity(), 1);
+  // Backing array and the strings. sizeof(Rep) = 2 * sizeof(int) for capacity
+  // and allocated_size. Each element of the Rep contributes sizeof(void*),
+  // and each string contributes sizeof(str) to arena memory.
+  EXPECT_EQ(
+      array->SpaceUsedExcludingSelf(),
+      2 * sizeof(int) + array->Capacity() * sizeof(void*) + 2 * sizeof(str));
   // We used some arena space now.
   EXPECT_LT(usage_before, arena.SpaceUsed());
   // And the pointer_begin is not in the sso anymore.
   EXPECT_FALSE(is_inlined());
 }
 
-TEST(RepeatedPtrField, CopyAssign) {
+TEST(RepeatedPtrFieldTest, CopyAssign) {
   RepeatedPtrField<std::string> source, destination;
   source.Add()->assign("4");
   source.Add()->assign("5");
@@ -758,7 +946,7 @@ TEST(RepeatedPtrField, CopyAssign) {
   EXPECT_EQ("5", destination.Get(1));
 }
 
-TEST(RepeatedPtrField, SelfAssign) {
+TEST(RepeatedPtrFieldTest, SelfAssign) {
   // Verify that assignment to self does not destroy data.
   RepeatedPtrField<std::string> source, *p;
   p = &source;
@@ -772,7 +960,7 @@ TEST(RepeatedPtrField, SelfAssign) {
   EXPECT_EQ("8", source.Get(1));
 }
 
-TEST(RepeatedPtrField, MoveConstruct) {
+TEST(RepeatedPtrFieldTest, MoveConstruct) {
   {
     RepeatedPtrField<std::string> source;
     *source.Add() = "1";
@@ -800,7 +988,7 @@ TEST(RepeatedPtrField, MoveConstruct) {
   }
 }
 
-TEST(RepeatedPtrField, MoveAssign) {
+TEST(RepeatedPtrFieldTest, MoveAssign) {
   {
     RepeatedPtrField<std::string> source;
     *source.Add() = "1";
@@ -896,7 +1084,7 @@ TEST(RepeatedPtrField, MoveAssign) {
   }
 }
 
-TEST(RepeatedPtrField, MutableDataIsMutable) {
+TEST(RepeatedPtrFieldTest, MutableDataIsMutable) {
   RepeatedPtrField<std::string> field;
   *field.Add() = "1";
   EXPECT_EQ("1", field.Get(0));
@@ -907,7 +1095,7 @@ TEST(RepeatedPtrField, MutableDataIsMutable) {
   EXPECT_EQ("2", field.Get(0));
 }
 
-TEST(RepeatedPtrField, SubscriptOperators) {
+TEST(RepeatedPtrFieldTest, SubscriptOperators) {
   RepeatedPtrField<std::string> field;
   *field.Add() = "1";
   EXPECT_EQ("1", field.Get(0));
@@ -917,7 +1105,7 @@ TEST(RepeatedPtrField, SubscriptOperators) {
   EXPECT_EQ(*field.data(), &const_field[0]);
 }
 
-TEST(RepeatedPtrField, ExtractSubrange) {
+TEST(RepeatedPtrFieldTest, ExtractSubrange) {
   // Exhaustively test every subrange in arrays of all sizes from 0 through 9
   // with 0 through 3 cleared elements at the end.
   for (int sz = 0; sz < 10; ++sz) {
@@ -937,7 +1125,6 @@ TEST(RepeatedPtrField, ExtractSubrange) {
           EXPECT_EQ(field.size(), sz + extra);
           for (int i = 0; i < extra; ++i) field.RemoveLast();
           EXPECT_EQ(field.size(), sz);
-          EXPECT_EQ(field.ClearedCount(), extra);
 
           // Create a catcher array and call ExtractSubrange.
           std::string* catcher[10];
@@ -959,9 +1146,7 @@ TEST(RepeatedPtrField, ExtractSubrange) {
             EXPECT_EQ(field.Mutable(i), subject[i + num]);
 
           // Reinstate the cleared elements.
-          EXPECT_EQ(field.ClearedCount(), extra);
           for (int i = 0; i < extra; ++i) field.Add();
-          EXPECT_EQ(field.ClearedCount(), 0);
           EXPECT_EQ(field.size(), sz - num + extra);
 
           // Make sure the extra elements are all there (in some order).
@@ -981,11 +1166,11 @@ TEST(RepeatedPtrField, ExtractSubrange) {
   }
 }
 
-TEST(RepeatedPtrField, DeleteSubrange) {
+TEST(RepeatedPtrFieldTest, DeleteSubrange) {
   // DeleteSubrange is a trivial extension of ExtendSubrange.
 }
 
-TEST(RepeatedPtrField, Cleanups) {
+TEST(RepeatedPtrFieldTest, Cleanups) {
   Arena arena;
   auto growth = internal::CleanupGrowth(
       arena, [&] { Arena::Create<RepeatedPtrField<std::string>>(&arena); });
@@ -997,6 +1182,37 @@ TEST(RepeatedPtrField, Cleanups) {
 }
 
 
+TEST(RepeatedPtrFieldDeathTest, CheckedGetOrAbortTest) {
+  RepeatedPtrField<std::string> field;
+
+  // Empty container tests.
+  EXPECT_DEATH(internal::CheckedGetOrAbort(field, -1), "index: -1, size: 0");
+  EXPECT_DEATH(internal::CheckedGetOrAbort(field, field.size()),
+               "index: 0, size: 0");
+
+  // Non-empty container tests
+  field.Add()->assign("foo");
+  field.Add()->assign("bar");
+  EXPECT_DEATH(internal::CheckedGetOrAbort(field, 2), "index: 2, size: 2");
+  EXPECT_DEATH(internal::CheckedGetOrAbort(field, -1), "index: -1, size: 2");
+}
+
+TEST(RepeatedPtrFieldDeathTest, CheckedMutableOrAbortTest) {
+  RepeatedPtrField<std::string> field;
+
+  // Empty container tests.
+  EXPECT_DEATH(internal::CheckedMutableOrAbort(&field, -1),
+               "index: -1, size: 0");
+  EXPECT_DEATH(internal::CheckedMutableOrAbort(&field, field.size()),
+               "index: 0, size: 0");
+
+  // Non-empty container tests
+  field.Add()->assign("foo");
+  field.Add()->assign("bar");
+  EXPECT_DEATH(internal::CheckedMutableOrAbort(&field, 2), "index: 2, size: 2");
+  EXPECT_DEATH(internal::CheckedMutableOrAbort(&field, -1),
+               "index: -1, size: 2");
+}
 // ===================================================================
 
 class RepeatedPtrFieldIteratorTest : public testing::Test {
@@ -1357,7 +1573,6 @@ TEST_F(RepeatedPtrFieldPtrsIteratorTest, PtrMutation) {
       proto_array_.pointer_begin();
   **iter = "moo";
   EXPECT_EQ("moo", proto_array_.Get(0));
-
   EXPECT_EQ("bar", proto_array_.Get(1));
   EXPECT_EQ("baz", proto_array_.Get(2));
   ++iter;
@@ -1567,14 +1782,14 @@ TEST_F(RepeatedPtrFieldInsertionIteratorsTest, MoveProtos) {
       RepeatedFieldBackInserter(testproto.mutable_repeated_nested_message()));
 
   ASSERT_EQ(src.size(), testproto.repeated_nested_message_size());
-  for (int i = 0; i < src.size(); ++i) {
+  for (size_t i = 0; i < src.size(); ++i) {
     EXPECT_EQ(src[i].DebugString(),
               testproto.repeated_nested_message(i).DebugString());
   }
 }
 
-}  // namespace
 
+}  // namespace internal
 }  // namespace protobuf
 }  // namespace google
 
